// <><><><><><><><><><><><><>  Plot3D.cpp  <><><><><><><><><><><><><><>
//
// ----------------------------------------------------
// (C) Copyright Clark Dailey, Inc. 2006
//
// Used as a base class for several Plot3D classes
//
// ----------------------------------------------------

#include "stdafx.h"	// precompiled header
#include "plot3d.h"
#include "surfx3d.h"
#include <sys/stat.h>
//#include "shlwapi.h"


// -------------------------------------------------------------------
//    I N I T I A L I Z A T I O N
// -------------------------------------------------------------------
Plot3D::Plot3D()
{
	InitBase();

	m_pWnd=0;
	m_pDC=0;
	m_pMemDC=0;
	m_hBitmap=0;
	m_wndWidth  = 1;
	m_wndHeight = 1;
	m_is_rendered  = 0;
	m_is_rendering = 0;
	m_is_equation_ok = 0;

}

// -------------------------------------------------------------------
Plot3D::~Plot3D()
{
	// release bitmap
	ReleaseBitmap();

	// release device context
	ReleaseDC();

	// no longer attached to a window
	m_pWnd = 0;
}

// ------------------------------------------------------------------------------------------
// set default user inputs
void Plot3D::DefInputsBase(void)
{
	// set axis ranges
	m_xmin=-10.;
	m_xmax =10.;
	m_ymin=-10.;
	m_ymax= 10.;
	m_zmin=-10.;
	m_zmax= 10.;

	// viewing angles
	m_xytilt=60.;
	m_zrotate=50.;

	// video resolution
	m_xvmin=0;
	m_yvmin=0;
	m_xvmax=640;
	m_yvmax=480;
	m_xvmid = (m_xvmin + m_xvmax)/2;
	m_yvmid = (m_yvmin + m_yvmax)/2;
	m_scale=90;

	// axis parameters
	m_show_axis=0;
	m_axis_color=DEF_AXIS_COLOR;
	memset(&m_axis,0,sizeof(m_axis));

	// color parameters
	m_is_color = 1;
	m_remove_hidden = 1;
	m_back_color=DEF_BKGND_COLOR;
	m_line_color=PLOT_COLOR_BLACK;
}

// ------------------------------------------------------------------------------------------
void Plot3D::ReleaseDC()
{
	m_is_rendered  = 0;

	// release device context
	if (m_pDC && m_pWnd) m_pWnd->ReleaseDC(m_pDC);
	m_pDC = 0;

	// release memory device context
	if (m_pMemDC) delete m_pMemDC;
	m_pMemDC = 0;
}

// ------------------------------------------------------------------------------------------
void Plot3D::ReleaseBitmap()
{
	// release bitmap
	if (m_hBitmap)
		DeleteObject(m_hBitmap);
	m_hBitmap = 0;
}

// ------------------------------------------------------------------------------------------
// input
//         pWnd
// outputs
//         m_xvmin, m_xvmax, m_yvmin, m_yvmax, m_xvmid, m_yvmid
// ------------------------------------------------------------------------------------------
void Plot3D::SetWnd(CWnd* pWnd)
{
	RECT rect;
	m_pWnd = pWnd;
	if (pWnd==NULL) return;
	pWnd->GetClientRect(&rect);

	// from window
	m_xvmin = rect.left;
	m_yvmin = rect.top;
	m_xvmax = rect.right;
	m_yvmax = rect.bottom;

	// calculated
	m_xvmid = (m_xvmin + m_xvmax)/2;
	m_yvmid = (m_yvmin + m_yvmax)/2;
}

// ------------------------------------------------------------------------------------------
//   C A L C U L A T I O N     R O U T I N E S
// ------------------------------------------------------------------------------------------

void Plot3D::InitBase()	// initialization
{
	DefInputsBase();

	// Dependencies
	CalcTM();
	CalcCorners();
	CalcAxis();

	PlotOutputOn();
	m_pending_moveto = 0;
	m_xv_cur = -1000;
	m_yv_cur = -1000;
	m_color_cur = 1;
	m_is_equation_ok = 0;
}

// ------------------------------------------------------------------------------------------
// calculate transformation matrix
// input
//       m_xytilt, m_zrotate;
// output
//       m_tm[][]
// ------------------------------------------------------------------------------------------
void Plot3D::CalcTM()
{
	double	zrotate,sin_zrotate,cos_zrotate;
	double	xytilt,sin_xytilt,cos_xytilt;

	zrotate = Deg2Rad(m_zrotate);
	xytilt  = Deg2Rad(m_xytilt);
	sin_zrotate = sin(zrotate);
	cos_zrotate = cos(zrotate);
	sin_xytilt  = sin(xytilt);
	cos_xytilt  = cos(xytilt);

	m_tm[0][0] =  cos_zrotate;
	m_tm[0][1] =  sin_zrotate;
	m_tm[0][2] =  0;
	m_tm[1][0] = -sin_zrotate * cos_xytilt;
	m_tm[1][1] =  cos_zrotate * cos_xytilt;
	m_tm[1][2] =                sin_xytilt;
	m_tm[2][0] =  sin_zrotate * sin_xytilt;
	m_tm[2][1] = -cos_zrotate * sin_xytilt;
	m_tm[2][2] =                cos_xytilt;
}

// ------------------------------------------------------------------------------------------
// calculate projections of corners
// inputs
//          m_xmin,  m_xmax,  m_ymin,  m_ymax,  m_zmin,  m_zmax
//          m_xvmin, m_xvmax, m_yvmin, m_yvmax
//          m_scale
//          m_tm[3][3]
//
// outputs
//          m_xcorner [2], m_ycorner [2], m_zcorner [2]
//          m_xpcorner[2], m_ypcorner[2], m_zpcorner[2]
//          m_xpbmid, m_ypbmid, m_zpbmid
//          m_p2v
// ------------------------------------------------------------------------------------------
void Plot3D::CalcCorners()
{
	double xc[2], yc[2], zc[2];
	double x,y,z,d;
	unsigned ix,iy,iz,ic;

	xc[0] = m_xmin;
	xc[1] = m_xmax;
	yc[0] = m_ymin;
	yc[1] = m_ymax;
	zc[0] = m_zmin;
	zc[1] = m_zmax;

	for (ic=0,iz=0;iz<2;iz++)
	{
		z = zc[iz];
		for (iy=0;iy<2;iy++)
		{
			y = yc[iy];
			for (ix=0;ix<2;ix++)
			{
				// corner values
				m_xcorner[ic] = x = xc[ix];
				m_ycorner[ic] = y;
				m_zcorner[ic] = z;
				// projection of corners
				m_xpcorner[ic] = xpf(x,y,z);
				m_ypcorner[ic] = ypf(x,y,z);
				m_zpcorner[ic] = zpf(x,y,z);
				ic++;
			} // for ix
		} // for iy
	} // for iz

	// find min,max plot coordinates 
	m_xpbmin = m_xpbmax = m_xpcorner[0];
	m_ypbmin = m_ypbmax = m_ypcorner[0];
	m_zpbmin = m_zpbmax = m_zpcorner[0];
	for (unsigned i=1;i<8;i++)
	{
		if (m_xpcorner[i]<m_xpbmin) m_xpbmin = m_xpcorner[i];
		if (m_xpcorner[i]>m_xpbmax) m_xpbmax = m_xpcorner[i];
		if (m_ypcorner[i]<m_ypbmin) m_ypbmin = m_ypcorner[i];
		if (m_ypcorner[i]>m_ypbmax) m_ypbmax = m_ypcorner[i];
		if (m_zpcorner[i]<m_zpbmin) m_zpbmin = m_zpcorner[i];
		if (m_zpcorner[i]>m_zpbmax) m_zpbmax = m_zpcorner[i];
	}
	m_xpbmid = (m_xpbmin + m_xpbmax)/2.;
	m_ypbmid = (m_ypbmin + m_ypbmax)/2.;
	m_zpbmid = (m_zpbmin + m_zpbmax)/2.;

	// calc plot to video scaling
	m_p2v = ((m_scale/100.) * (double)(m_xvmax - m_xvmin)) / (double)(m_xpbmax - m_xpbmin);
	d     = ((m_scale/100.) * (double)(m_yvmax - m_yvmin)) / (double)(m_ypbmax - m_ypbmin);
	if (d < m_p2v)
		m_p2v = d;

	// plotting must start from outmost of the screen inward for proper hidden line removal
	if (m_zpcorner[1]>=m_zpcorner[0])
	{
		m_xstart[0] =  m_xmax;
		m_xstart[1] =  m_xmin;
	}
	else
	{
		m_xstart[0] =  m_xmin;
		m_xstart[1] =  m_xmax;
	}
	if (m_zpcorner[2]>m_zpcorner[0])
	{
		m_ystart[0] =  m_ymax;
		m_ystart[1] =  m_ymin;
	}
	else
	{
		m_ystart[0] =  m_ymin;
		m_ystart[1] =  m_ymax;
	}
	if (m_ypcorner[0]<m_ypcorner[4])
	{
		m_zstart[0] = m_zmin;
		m_zstart[1] = m_zmax;
	}
	else
	{
		m_zstart[0] = m_zmax;
		m_zstart[1] = m_zmin;
	}
}

// ------------------------------------------------------------------------------------------
// calculate axis info
void Plot3D::CalcAxis()
{
	memset((char*)&m_axis,0,sizeof(m_axis));
	m_axis.nxticks = 3;
	m_axis.nyticks = 3;
	m_axis.nzticks = 3;
	strcpy(m_axis.font_name,"Arial"); // needs to be TrueType
	m_axis.font_height  = 12;
	m_axis.font_width   = 7;
	m_axis.tick_length  = 20;
	m_axis.tick_spacing = 20;
}

// ------------------------------------------------------------------------------------------
// calculate centroid of cube
void Plot3D::GetCenter(double* xmid,double* ymid,double* zmid)
{
	*xmid = (m_xmin + m_xmax)/2.;
	*ymid = (m_ymin + m_ymax)/2.;
	*zmid = (m_zmin + m_zmax)/2.;
}


// -------------------------------------------------------------------
//     P L O T T I N G     P R I M I T I V E S
// -------------------------------------------------------------------
void Plot3D::WinMoveTo(int xv,int yv)
{
	// draw on screen
	m_pDC->MoveTo(xv,yv);		// issue a move to windows

	// echo to memory DC
	m_pMemDC->MoveTo(xv,yv);	// issue a move to windows
}

// -------------------------------------------------------------------
void Plot3D::WinDrawTo(int xv,int yv)
{
	CPen Pen(PS_SOLID,1,m_line_color);	// construct a pen with the desired color

	// draw on screen
	CPen* pOldPen1 = m_pDC->SelectObject(&Pen);		// select the pen into the device context
	m_pDC->LineTo(xv,yv);				   			// draw the colored line
	m_pDC->SelectObject(pOldPen1);					// restore the original pen

	// echo to memory DC
	CPen* pOldPen2 = m_pMemDC->SelectObject(&Pen);	// select the pen into the device context
	m_pMemDC->LineTo(xv,yv);					   	// draw the colored line
	m_pMemDC->SelectObject(pOldPen2);				// restore the original pen
}

// -------------------------------------------------------------------
void Plot3D::WinColorTo(long color)
{
	m_line_color = color;	// just save it
}

// -------------------------------------------------------------------
void Plot3D::WinSetPixelColor(int xv,int yv,long color)
{
	// draw on screen
	m_pDC->SetPixel(xv,yv,color);

	// echo to memory DC
	m_pMemDC->SetPixel(xv,yv,color);
}

// -------------------------------------------------------------------
void Plot3D::WinPlotText(AXIS_PLOT_INFO* pAxisInfo,double degrees,int xv,int yv,CString text)
{
	BOOL bRC;
	double radians;

	if (m_pMemDC)
	{
		// setup font info
		LOGFONT logFont;
		memset((char*)&logFont,0,sizeof(logFont));
		strcpy(logFont.lfFaceName,pAxisInfo->font_name);
		logFont.lfHeight = pAxisInfo->font_height;
		logFont.lfWidth  = pAxisInfo->font_width;

		// check if need to invert text (don't draw text upside down)
		radians = Deg2Rad(degrees);
		if (cos(radians)<0.)
		{
			// need to invert text
			CFont extFont;

			// determine width and height of text at zero degrees
			bRC = extFont.CreateFontIndirect(&logFont);
			m_pMemDC->SelectObject(extFont);
			CSize textExt = m_pMemDC->GetTextExtent(text);
			
			// calculate opposite corner
			xv = (int)((double)xv + (double)(textExt.cx)*cos(radians) - (double)(textExt.cy)*sin(radians));
			yv = (int)((double)yv - (double)(textExt.cx)*sin(radians) - (double)(textExt.cy)*cos(radians));
			degrees += 180.; // add 180 degrees
			if (degrees > 360.) degrees -= 360.;
		}

		// create a rotated font
		CFont rotFont;
		logFont.lfEscapement = (long)(degrees*10.);	// tenths of degree
		bRC = rotFont.CreateFontIndirect(&logFont);

		// output to screen
		m_pDC->SelectObject(rotFont);
		m_pDC->TextOut(xv,yv,text);

		// echo to memory DC
		m_pMemDC->SelectObject(rotFont);
		m_pMemDC->TextOut(xv,yv,text);
	}
}

// ------------------------------------------------------------------------------------------
void Plot3D::MapAndDrawLine(double x1,double y1,double z1,double x2,double y2,double z2)
{
	int	xv1,yv1,xv2,yv2;

	// Map points
	MapPoint(x1,y1,z1,&xv1,&yv1);
	MapPoint(x2,y2,z2,&xv2,&yv2);

	// draw the line
	FilterMoveTo(xv1,yv1);
	FilterDrawTo(xv2,yv2);
}

// ------------------------------------------------------------------------------------------
// filter move commands
void Plot3D::FilterMoveTo(int xv,int yv)
{
	if (m_xv_cur==xv && m_yv_cur==yv) return; // already there
	if (m_to_plotter)
		WinMoveTo(xv,yv);
	else
		m_pending_moveto = 1;
	m_xv_cur = xv;
	m_yv_cur = yv;
}

// ------------------------------------------------------------------------------------------
// filter draw commands
void Plot3D::FilterDrawTo(int xv,int yv)
{
	if (m_xv_cur==xv && m_yv_cur==yv) return; // already there
	if (m_to_plotter)
	{
		if (m_pending_moveto) WinMoveTo(m_xv_cur,m_yv_cur);
		m_pending_moveto = 0;
		WinDrawTo(xv,yv);
	}
	m_xv_cur = xv;
	m_yv_cur = yv;
}

// ------------------------------------------------------------------------------------------
void Plot3D::FilterColorTo(long color)
{
	if (color==m_color_cur) return; // already got this color
	if (m_to_plotter) WinColorTo(color);
	m_color_cur = color;
}

// ------------------------------------------------------------------------------------------
void Plot3D::ClearScreen(void)
{
	if (m_pWnd == NULL) return;
	ReleaseBitmap();
	ReleaseDC();
	m_pWnd->Invalidate();
}

// -------------------------------------------------------------------
void Plot3D::OnPaint(CDC* pDC)
{
	if (m_pWnd)
	{
		// paint background color first
		CRect rect;
		m_pWnd->GetClientRect(&rect);
		pDC->FillSolidRect(&rect,m_back_color);
	}
	CopyPlotToDC(pDC);
}

// -------------------------------------------------------------------
void Plot3D::WinBegPlot()
{
	RECT	rect;

	// release any device contexts
	ReleaseBitmap();
	ReleaseDC();
	theApp.UpdateToolBar();

	// get window size
	m_abort_plotting = 0;
	if (!m_pWnd) return;
	m_pWnd->GetClientRect(&rect);
	m_wndWidth  = rect.right  - rect.left;
	m_wndHeight = rect.bottom - rect.top ;
	if (m_wndWidth <1) m_wndWidth=1;				       
	if (m_wndHeight<1) m_wndHeight = 1;
	CSize wsize = CSize(m_wndWidth,m_wndHeight);

	// get device context
	m_pDC = m_pWnd->GetDC();
	if (m_pDC==0) return;

	// create a memory device context
	m_pMemDC = new CDC;
	if (m_pMemDC==0) return;
	m_pMemDC->CreateCompatibleDC(m_pDC);
	m_hBitmap = CreateCompatibleBitmap(m_pDC->m_hDC, m_wndWidth, m_wndHeight);
	SelectObject(m_pMemDC->m_hDC, m_hBitmap);

	CPen Pen(PS_SOLID,1,m_line_color);	// construct a pen with the desired color

	// screen DC
	m_pDC->SetBkColor(m_back_color);
	m_pDC->SetMapMode(MM_ISOTROPIC);
	m_pDC->SetWindowOrg(0,wsize.cy);
	m_pDC->SetWindowExt(wsize);
	m_pDC->SetViewportExt(wsize);
	m_pDC->ScaleViewportExt(1,1,-1,1);			// invert y coordinate
	m_pDC->FillSolidRect(&rect,m_back_color);	// clear the screen
	m_pDC->SelectObject(&Pen);					// select the pen into the device context

	// memory DC
	m_pMemDC->SetBkColor(m_back_color);
	m_pMemDC->SetMapMode(MM_ISOTROPIC);
	m_pMemDC->SetWindowOrg(0,wsize.cy);
	m_pMemDC->SetWindowExt(wsize);
	m_pMemDC->SetViewportExt(wsize);
	m_pMemDC->ScaleViewportExt(1,1,-1,1);			// invert y coordinate
	m_pMemDC->FillSolidRect(&rect,m_back_color);	// clear the screen
	m_pMemDC->SelectObject(&Pen);					// select the pen into the device context

	m_is_rendering = 1;
	theApp.SetStatusString("Render Begin");
	theApp.UpdateToolBar();
}

// -------------------------------------------------------------------
void Plot3D::WinEndPlot()
{
	// release device context
	if (m_pWnd && m_pDC)
		m_pWnd->ReleaseDC(m_pDC);
	m_pDC = 0;
	m_is_rendering = 0;
	m_is_rendered  = 1;
	theApp.SetStatusString("Render Complete");
	theApp.UpdateToolBar();
}

// -------------------------------------------------------------------
// return difference between two times
void Plot3D::CalcDiffTime(time_t begTime,time_t endTime,
				  unsigned* minutes,unsigned* seconds)
{
	unsigned m, s;
	s = (unsigned)difftime(endTime,begTime); // total seconds
	m = s/60;
	s = s - m*60;
	*minutes = m;
	*seconds = s;
}

// --------------------------------------------------------------
// copy plot memory mapped image to device context
void Plot3D::CopyPlotToDC(CDC* pDC)
{
	double s1, s2;
	int pw, ph, sw, sh=0, px, py;
	int scale = 1;

	if (!m_pMemDC) return;

	pw = m_wndWidth;
	ph = m_wndHeight;
	if (pDC->IsPrinting())
	{
		// printer width and height
		pw = pDC->GetDeviceCaps(HORZRES);  // pixels
		ph = pDC->GetDeviceCaps(VERTRES);  // pixels

		if (pw<1) pw = 1;
		if (ph<1) ph = 1;

		// screen width and height
		sw = m_wndWidth;
		sh = m_wndHeight;
		if (sw<1) sw = 1;
		if (sh<1) sh = 1;

		// calculate x and y scales; use plot scale 
		s1 = (pw / sw) * (m_scale/100.);
		s2 = (ph / sh) * (m_scale/100.);

		// take smaller of scales
		scale = (int)min(s1,s2);
		if (scale < 1) scale = 1;
	}

	CSize wsize = CSize(m_wndWidth,m_wndHeight);
	pDC->SetBkColor(m_back_color);
	pDC->SetMapMode(MM_ISOTROPIC);
	pDC->SetWindowOrg(0,wsize.cy);
	pDC->SetWindowExt(wsize);
	pDC->SetViewportExt(wsize);
	pDC->ScaleViewportExt(scale,1,-scale,1);	// invert y coordinate
	pDC->SetStretchBltMode(STRETCH_DELETESCANS); 
	if (m_pMemDC==NULL) return;

	if (pDC->IsPrinting())
	{
		// MSDN Note: Q195830 INFO: Blitting Between DCs for Different Devices Is Unsupported
		// use Device Independent Bitmap (DIB) for printer
		BITMAPINFO bmInfo;
		HANDLE hRAM;
		char* pBits;
	
		// build bitmap info structure
		int imageSize = ((m_wndWidth*3+3)&0xFFFFFFFC)*m_wndHeight;
		memset(&bmInfo,0,sizeof(bmInfo));
		bmInfo.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
		bmInfo.bmiHeader.biWidth  = m_wndWidth;
		bmInfo.bmiHeader.biHeight = m_wndHeight;
		bmInfo.bmiHeader.biPlanes = 1; 
		bmInfo.bmiHeader.biBitCount = 24;
		bmInfo.bmiHeader.biCompression = BI_RGB;
		bmInfo.bmiHeader.biSizeImage = imageSize; 
	
		// allocate memory for bitmap RGB bits
		hRAM = ::GlobalAlloc(GMEM_FIXED,imageSize);
		pBits = (char*)hRAM;
		if (pBits)
		{
			// get the DIB bits
			GetDIBits(m_pMemDC->m_hDC,m_hBitmap,0,m_wndHeight,pBits,&bmInfo,DIB_RGB_COLORS);
		
			// stretch to correct size
			bmInfo.bmiHeader.biHeight = -m_wndHeight; // DIBs are upside down
			sw = m_wndWidth  * scale;
			sh = m_wndHeight * scale;
			px = pw > sw ? (pw-sw)/2 : 0;
			py = ph > sh ? (ph-sh)/2 : 0;
			px =  px / scale;
			py = -py / scale;
			::StretchDIBits(pDC->m_hDC, px, py, m_wndWidth, m_wndHeight,
							            0,  0,  m_wndWidth, m_wndHeight,	  
										pBits,  &bmInfo, DIB_RGB_COLORS, SRCCOPY);
			// delete memory
			::GlobalFree(hRAM);
		}
	}
	else
	{
		pDC->StretchBlt(0,0,m_wndWidth,m_wndHeight,m_pMemDC,0,0,m_wndWidth,m_wndHeight,SRCCOPY);
	}
}

// -------------------------------------------------------------------
// returns: 0=ok, 1=no copy
int Plot3D::CopyPlotToClipBoard()
{
	BOOL bval;
	int rc = 0;

	if (!m_hBitmap) return(1);

	// Open the clipboard
	if (OpenClipboard(NULL))
	{
		// Clean the contents of the clipboard
		bval = EmptyClipboard();

	 	// ----- put data to the clipboard data -----
	 	SetClipboardData(CF_BITMAP, m_hBitmap);

  		// Close the clipboard
	 	bval = CloseClipboard();
	}
	else
	{
		AfxMessageBox( "Clipboard unavailable" );
		rc = 1;
	}
	return(rc);
}

// -------------------------------------------------------------------
// returns: 0=no, 1=user cancelled plot
int Plot3D::IsAborted()
{
	if (m_abort_plotting) return(1);
	// keep window messages pumping to allow plot redraw
	MsgPump();
	return(0);
}

// -------------------------------------------------------------------
// keep window messages pumping to allow plot redraw
void Plot3D::MsgPump()
{
	MSG msg;
	while ( ::PeekMessage( &msg, NULL, 0, 0, PM_NOREMOVE ) ) 
	{ 
		if ( !AfxGetApp()->PumpMessage( ) ) 
		{ 
			::PostQuitMessage(0); 
		} 
	}
}

// -------------------------------------------------------------------
//     A X I S     R O U T I N E S
// -------------------------------------------------------------------

void Plot3D::DrawAxis(void)
{
	CString str;
	double dx,dy,dz,x1,y1,z1,x2,y2,xp,yp;
	double degrees,gap_length,tick_length;
	int xv1,yv1,xv2,yv2;
	unsigned n;

	if (!m_show_axis) return;
	if (m_abort_plotting) return;

	FilterColorTo(m_axis_color);

	// Constant Y Front Edge
	MapAndDrawLine(m_xstart[0],m_ystart[0],m_zstart[0],
				   m_xstart[1],m_ystart[0],m_zstart[0]);

	// Constant X Front Edge
	MapAndDrawLine(m_xstart[0],m_ystart[0],m_zstart[0],
				   m_xstart[0],m_ystart[1],m_zstart[0]);

	// Vertical Front Edges
	MapAndDrawLine(m_xstart[0],m_ystart[0],m_zstart[0],
				   m_xstart[0],m_ystart[0],m_zstart[1]);

	MapAndDrawLine(m_xstart[1],m_ystart[0],m_zstart[0],
				   m_xstart[1],m_ystart[0],m_zstart[1]);

	MapAndDrawLine(m_xstart[0],m_ystart[1],m_zstart[0],
				   m_xstart[0],m_ystart[1],m_zstart[1]);

	// -------------
	// X tick marks
	// -------------
	if (m_axis.nxticks > 0)
	{
		n = m_axis.nxticks-1;
		if (n==0) n=1; // avoid divions by zero
		dx =  m_xstart[0] - m_xstart[1];
		dy = (m_ystart[1] - m_ystart[0])/((double)n);
		gap_length  = dx * (double)m_axis.tick_spacing / 500.;
		tick_length = dx * (double)m_axis.tick_length  / 500.;
	
		x1 = m_xstart[0]+gap_length;  y1 = m_ystart[0];	z1 = m_zstart[0];
		x2 = x1 + tick_length;
		MapPoint(m_xstart[1],m_ystart[0],m_zstart[0],&xv1,&yv1);
		MapPoint(m_xstart[0],m_ystart[0],m_zstart[0],&xv2,&yv2);
		degrees = Rad2Deg( atan2( -(double)(yv2-yv1), (double)(xv2-xv1) ) );
		for (n=0;n<m_axis.nxticks;n++,y1+=dy)
		{
			MapPoint(x1,y1,z1,&xv1,&yv1);
			MapPoint(x2,y1,z1,&xv2,&yv2);
			FilterMoveTo(xv1,yv1);
			FilterDrawTo(xv2,yv2);
			str.Format("y=%-8.4lf",y1);
			WinPlotText(&m_axis,degrees,xv2,yv2,str);
		}
	}

	// -------------
	// Y tick marks
	// -------------
	if (m_axis.nyticks > 0)
	{
		n = m_axis.nyticks-1;
		if (n==0) n=1; // avoid divions by zero
		dy =  m_ystart[0] - m_ystart[1];
		dx = (m_xstart[1] - m_xstart[0])/((double)n);
		gap_length  = dy * (double)m_axis.tick_spacing / 500.;
		tick_length = dy * (double)m_axis.tick_length  / 500.;
	
		y1 = m_ystart[0]+gap_length;  x1 = m_xstart[0];	z1 = m_zstart[0];
		y2 = y1 + tick_length;
		MapPoint(m_xstart[0],m_ystart[1],m_zstart[0],&xv1,&yv1);
		MapPoint(m_xstart[0],m_ystart[0],m_zstart[0],&xv2,&yv2);
		degrees = Rad2Deg( atan2( -(double)(yv2-yv1), (double)(xv2-xv1) ) );
		for (n=0;n<m_axis.nyticks;n++,x1+=dx)
		{
			MapPoint(x1,y1,z1,&xv1,&yv1);
			MapPoint(x1,y2,z1,&xv2,&yv2);
			FilterMoveTo(xv1,yv1);
			FilterDrawTo(xv2,yv2);
			str.Format("x=%-8.4lf",x1);
			WinPlotText(&m_axis,degrees,xv2,yv2,str);
		}
	}

	// -------------
	// Z tick marks
	// -------------
	if (m_axis.nzticks > 0)
	{
		n = m_axis.nzticks-1;
		if (n==0) n=1; // avoid divions by zero
		dz =  m_zstart[1] - m_zstart[0];
		gap_length  = dz * (double)m_axis.tick_spacing / 500.;
		tick_length = dz * (double)m_axis.tick_length  / 500.;
		dz = dz/(double)n;

		x1 = m_xcorner[0];  y1 = m_ycorner[0];  z1=m_xpcorner[0];
		if (m_xpcorner[1]>z1)
		{	x1 = m_xcorner[1];  y1 = m_ycorner[1];  z1=m_xpcorner[1];	}
		if (m_xpcorner[2]>z1)
		{	x1 = m_xcorner[2];  y1 = m_ycorner[2];  z1=m_xpcorner[2];	}
		if (m_xpcorner[3]>z1)
		{	x1 = m_xcorner[3];  y1 = m_ycorner[3];  z1=m_xpcorner[3];	}
		MapAndDrawLine(x1,y1,m_zstart[0],
					   x1,y1,m_zstart[1]);
		z1 = m_zstart[0];
		for (n=0;n<m_axis.nzticks;n++,z1+=dz)
		{
			xp = xpf(x1,y1,z1)+gap_length;
			yp = ypf(x1,y1,z1);
			xv1 = xvf(xp);
			yv1 = yvf(yp);
			xp += tick_length;
			xv2 = xvf(xp);
			yv2 = yvf(yp);
			FilterMoveTo(xv1,yv1);
			FilterDrawTo(xv2,yv2);
			str.Format("z=%-8.4lf",z1);
			WinPlotText(&m_axis,0.,xv2,yv2+m_axis.font_height,str);
		}
	}
}

// -------------------------------------------------------------------
//     U T I L I T Y     R O U T I N E S
// -------------------------------------------------------------------

// returns: 0=no, 1=ini plot file exists
int DoesFileExist(CString sFilePath)
{
	struct _stat fileInfo;
	int rc = ::_stat(sFilePath,&fileInfo);
	return(rc==0?1:0);
//	return( PathFileExists(sFilePath) );
}

// --------------------------------------------------------------
// extract folder path (without ending backslash) from fully qualified filename
CString ExtractPathDir(CString pathName)
{
	CString sDir, fn=pathName;
	int indx = fn.ReverseFind('\\');  // find last backslash
	if (indx>0) sDir = fn.Left(indx);
	return(sDir);
}

// --------------------------------------------------------------
// extract filename (no extension) from fully qualified filename
CString ExtractFileName(CString fullName)
{
	CString sName = fullName;

	// find file extension
	int indx = sName.ReverseFind('.');
	if (indx>=0) sName = sName.Left(indx);	// strip extension

	// find end of folder name
	indx = sName.ReverseFind('\\');
	if (indx>=0)
		sName = sName.Mid(indx+1);	// extract filename
	else
	{
		// check for forward slash
		indx = sName.ReverseFind('/');
		if (indx>=0)
			sName = sName.Mid(indx+1);	// extract filename
	}
	return(sName);
}

// -------------------------------------------------------------------
//     I N I    F I L E     R O U T I N E S
// -------------------------------------------------------------------
CString Plot3D::GetIniStr(CString keyName,CString defValue)
{
	char buffer[256];
	CString value = defValue;
	memset(buffer,0,sizeof(buffer));
	DWORD cnt = ::GetPrivateProfileString(m_sIniHdr, keyName, defValue, buffer, sizeof(buffer), m_FilePath);
	if (cnt) value = buffer;
	return(value);
}

// -------------------------------------------------------------------
double Plot3D::GetIniDouble(CString keyName,double defValue)
{
	double value = defValue;
	char buffer[256];
	CString defStr;
	defStr.Format("%lf", defValue);
	memset(buffer,0,sizeof(buffer));
	DWORD cnt = ::GetPrivateProfileString(m_sIniHdr, keyName, defStr, buffer, sizeof(buffer), m_FilePath);
	if (cnt)
	{
		double dtemp;
		int rc=sscanf(buffer,"%lf", &dtemp);
		if (rc==1) value = dtemp;
	}
	return(value);
}

// -------------------------------------------------------------------
long Plot3D::GetIniLong(CString keyName,long defValue)
{
	long value = defValue;
	value = ::GetPrivateProfileInt(m_sIniHdr, keyName, defValue, m_FilePath);
	return(value);
}

// -------------------------------------------------------------------
// returns: 0=ok
int Plot3D::SetIniStr(CString keyName,CString value)
{
	BOOL rc = ::WritePrivateProfileString(m_sIniHdr, keyName, value, m_FilePath);
	return(rc?0:1);
}

// -------------------------------------------------------------------
// returns: 0=ok
int Plot3D::SetIniDouble(CString keyName,double value)
{
	CString strValue;
	strValue.Format("%lf", value);
	BOOL rc = ::WritePrivateProfileString(m_sIniHdr, keyName, strValue, m_FilePath);
	return(rc?0:1);
}

// -------------------------------------------------------------------
// returns: 0=ok
int Plot3D::SetIniLong(CString keyName,long value)
{
	CString strValue;
	strValue.Format("%ld", value);
	BOOL rc = ::WritePrivateProfileString(m_sIniHdr, keyName, strValue, m_FilePath);
	return(rc?0:1);
}

/*	<><><><><><><><><><><><><>  Plot3D.cpp  <><><><><><><><><><><><><><> */



