// <><><><><><><><><><><><><>  Plot3D.h  <><><><><><><><><><><><><><>
//
// ----------------------------------------------------
// (C) Copyright Clark Dailey, Inc. 2006
//
// Used as a base class for several Plot3D classes
//
// ----------------------------------------------------

#ifndef	_PLOT3D_H_	 	// avoid repeated including
#define	_PLOT3D_H_	1	// has been included

#include <math.h>
#include <assert.h>
#include "plotcalc.h"

#ifdef _DEBUG
 #define PLOT_ASSERT(v)	assert(v)
#else
 #define PLOT_ASSERT(v)	// do nothing
#endif

// degrees/radians conversions
inline double Deg2Rad(double deg) { return((deg*2.*3.1415925359)/360.);   }
inline double Rad2Deg(double rad) { return((rad*360.)/(2.*3.1415925359)); }
#define	PLOT3D_DBL_MAX	(1.e300)	// largest double value

// color values
#define	PLOT3D_MIN_COLOR	(0)
#define	PLOT3D_MAX_COLOR	(100)
#define	WIN_MIN_COLOR		(0)
#define	WIN_MAX_COLOR		(255)

// convert plot color component to windows color component
#define	PCOLOR_TO_WINCOLOR(d)		((int)((double)d*(double)WIN_MAX_COLOR/(double)PLOT3D_MAX_COLOR))

// ----------
// axis info
// ----------
typedef struct axis_s
{
	unsigned nxticks;		// number of x tick marks
	unsigned nyticks;		// number of y tick marks
	unsigned nzticks;		// number of z tick marks
	char	 font_name[128];// name of font (needs to be True-Type for rotating)
	int		 font_height;	// pixels
	int		 font_width;	// pixels
	unsigned tick_length;	// 1-100
	unsigned tick_spacing;	// 1-100
} AXIS_PLOT_INFO;

// defined colors
#define  PLOT_COLOR_BLACK	RGB(  0,  0,  0)	// black color
#define  PLOT_COLOR_WHITE	RGB(255,255,255)	// white color
#define  PLOT_COLOR_LTGRAY	RGB(212,208,200)	// gray background color

// default color mappings
#define	 DEF_LINE_COLOR		PLOT_COLOR_BLACK	// default line color
#define	 DEF_AXIS_COLOR		PLOT_COLOR_BLACK	// default axis color
#define	 DEF_BKGND_COLOR 	PLOT_COLOR_LTGRAY	// default background color

// ------------------------------------------------------------------------------------------
// base class for several Plot3D classes
class Plot3D
{
public:
	Plot3D();
	virtual ~Plot3D();

	// set routines
	void SetFilePath(CString filePath) { m_FilePath = filePath; }
	void RequestCancel(void) { m_abort_plotting = 1; }

	// status routines
	int IsRendering() { return(m_is_rendering); }
	int IsRendered()  { return(m_is_rendered);  }

	// paint plot image
	void OnPaint(CDC* dc);
	void ClearScreen(void);
	void CalcDiffTime(time_t begTime,time_t endTime,unsigned* minutes,unsigned* seconds);
	int  CopyPlotToClipBoard();
	void CopyPlotToDC(CDC* pDC);
	void MsgPump(void);
	void DrawAxis(void);

public:	// user settings

	// get mid coordinate range values
	void GetCenter(double* xmid,double* ymid,double* zmid);

protected:	// routines

	void SetWnd(CWnd* pWnd=0);
	void DefInputsBase(void);	// set default inputs
	void InitBase(void);		// initialization

	// validate color value
	double ValidateColorValue(double c)
	{
		if (c < PLOT3D_MIN_COLOR) return(PLOT3D_MIN_COLOR);
		if (c > PLOT3D_MAX_COLOR) return(PLOT3D_MAX_COLOR);
		return(c);
	}

	// INI file input
	void SetIniHeader(CString iniHdr) { m_sIniHdr=iniHdr; }
	CString GetIniStr   (CString keyName, CString defValue);
	double  GetIniDouble(CString keyName, double  defValue);
	long    GetIniLong  (CString keyName, long    defValue);

	// INI file output
	int     SetIniStr   (CString keyName, CString value);
	int     SetIniDouble(CString keyName, double  value);
	int     SetIniLong  (CString keyName, long    value);

	// ------------------------------------------------------------
	//             Transformation Equations
	// ------------------------------------------------------------
	// object space to plot space
	double xpf(double xo,double yo,double zo) { return(m_tm[0][0]*xo + m_tm[0][1]*yo + m_tm[0][2]*zo); }
	double ypf(double xo,double yo,double zo) { return(m_tm[1][0]*xo + m_tm[1][1]*yo + m_tm[1][2]*zo); }
	double zpf(double xo,double yo,double zo) { return(m_tm[2][0]*xo + m_tm[2][1]*yo + m_tm[2][2]*zo); } 
	
	// plot space to object space
	double xof(double xp,double yp,double zp) { return(m_tm[0][0]*xp + m_tm[1][0]*yp + m_tm[2][0]*zp); }
	double yof(double xp,double yp,double zp) { return(m_tm[0][1]*xp + m_tm[1][1]*yp + m_tm[2][1]*zp); }
	double zof(double xp,double yp,double zp) { return(m_tm[0][2]*xp + m_tm[1][2]*yp + m_tm[2][2]*zp); }
	
	// plot space to video space
	int xvf(double xp) { return( (int)((m_p2v*(xp-m_xpbmid) + (double)m_xvmid)) ); }
	int yvf(double yp) { return( (int)((m_p2v*(yp-m_ypbmid) + (double)m_yvmid)) ); }
	void MapPoint(double x,double y,double z,int* xv,int* yv) { *xv = xvf( xpf( x, y, z) ); *yv = yvf( ypf( x, y, z) ); }
	void MapAndDrawLine(double x1,double y1,double z1,double x2,double y2,double z2);

	// video space to plot space
	double xv2xp(int xv) { return( (double)(xv-m_xvmid)/m_p2v + m_xpbmid ); }
	double yv2yp(int yv) { return( (double)(yv-m_yvmid)/m_p2v + m_ypbmid ); }

	// ouput control
	void PlotOutputOff() { m_to_plotter=0; }
	void PlotOutputOn()  { m_to_plotter=1; }

	// filter move/draw commands
	void FilterMoveTo(int xv,int yv);
	void FilterDrawTo(int xv,int yv);
	void FilterColorTo(long color);

	// calculation routines
	void CalcCorners(void);	// calculate projections of corners
	void CalcAxis(void);	// calculate axis info
	void CalcTM(void);		// calculate transformation matrix

	// supply primitive move/draw for Windows
	void WinMoveTo(int xv,int yv);
	void WinDrawTo(int xv,int yv);
	void WinColorTo(long color);
	void WinSetPixelColor(int xv,int yv,long color);
	void WinPlotText(AXIS_PLOT_INFO* pAxisInfo,double degrees,int xv,int yv,CString text);

	// begin / end plot routines
	void WinBegPlot(void);
	void WinEndPlot(void);

	// check for user abort
	int  IsAborted(void);

	// color
	long GetBackColor() { return(m_back_color); }

private:	// internal routines
	void ReleaseDC(void);
	void ReleaseBitmap(void);

	// ------
	//  Data
	// ------
protected: // user inputs
	// x,y,z min/max ranges
	double	m_xmin, m_xmax, m_ymin, m_ymax, m_zmin, m_zmax;

	// viewing angles
	double	m_xytilt, m_zrotate;	// degrees

	// color info
	COLORREF m_line_color;
	COLORREF m_back_color;
	COLORREF m_axis_color;
	int		 m_is_color;
	int      m_remove_hidden;		// 0=no, 1=remove hidden lines

	// x,y video min/max ranges   SetWnd() sets these
	int m_xvmin, m_xvmax, m_yvmin, m_yvmax;
	int m_xvmid, m_yvmid;	// mid points of video coordinate system

	// plotting scale
	double	m_scale;

	// axis parameters
	BOOL  m_show_axis;  // 0=no, 1=draw axis

protected: // set
	// current pen location
	int	 m_xv_cur, m_yv_cur;

	// maps plot coordinates to video coordinates
	double m_p2v;

protected: // calculated

	// min/mid/max projection values (base plot)
	double m_xpbmin, m_xpbmid,  m_xpbmax;
	double m_ypbmin, m_ypbmid,  m_ypbmax;
	double m_zpbmin, m_zpbmid,  m_zpbmax;

	// axis info
	AXIS_PLOT_INFO m_axis;

	// control flags
	int	m_abort_plotting;	// 0=no, 1=abort plotting
	int m_is_equation_ok;	// 0=no, 1=plotting equation syntax is ok

	// status message
	CString m_sEquationStatus;  // equation status string

	// corner coordinates
	double m_xcorner [8],m_ycorner [8],m_zcorner [8];	// corner values
	double m_xpcorner[8],m_ypcorner[8],m_zpcorner[8];	// projection of corners
	double m_xstart  [2],m_ystart  [2],m_zstart  [2];	// starting x,y,z points: 0=even,1=odd

	CString m_FilePath;	// fully qualified file path

private:	// internal variables

	CString m_sIniHdr;	// INI section header

	// transformation matrix
	double m_tm[3][3];

	// output variables
	int m_to_plotter;	 	// 0=no, 1=output to plotter
	int m_pending_moveto;	// 0=no, 1=need to move before drawing

	// current pen color
	long m_color_cur;

	// windows drawing
	CWnd* m_pWnd;		// pointer to our window (we don't own this resource)
	CDC*  m_pMemDC;		// memory Device Context
	CDC*  m_pDC;		// device context
	HBITMAP m_hBitmap; 	// bitmap for drawing

	BOOL m_is_rendering;// 0=no, 1=rendering image
	BOOL m_is_rendered; // 0=no, 1=image has been rendered
	int  m_wndWidth;	// drawing window width  (pixels)
	int  m_wndHeight;	// drawing window height (pixels)
};

// Prototyping
int DoesFileExist(CString sFilePath);
CString ExtractPathDir(CString pathName);
CString ExtractFileName(CString fullName);


#endif	/* if _PLOT3D_H_ included */

/*	<><><><><><><><><><><><><>  Plot3D.h  <><><><><><><><><><><><><><> */



